/*  DB8085.C     ( Disassemble Binary 8085 )
***********************************************************************
*   Disassemble 8085 code from a binary file.
*   Two passes:
*     1st - builds label file for forward references
*     2nd - builds source file including forward references
*   References to addresses of instruction operands are noted
*   as sync errors.
*
*   Usage:  DB8085  binfile  [hexoffset]
*
*   Original written by Kevin D. Quitt  30 Nov 1988
*   Updated/modified by Richard M. Bender  13 April 1999
***********************************************************************/
#include    <stdio.h>
#include    <string.h>

#define MAX_LABELS  2048

unsigned int    Labels[ MAX_LABELS ];
unsigned char   In_File[ 132 ];
unsigned char   Out_File[ 132 ];
unsigned char   Temp[ 132 ];

char    *Op_Code[256]   =
   {
/*    0           1           2           3                             */
    "NOP",      "LXI\tB,",  "STAX\tB",  "INX\tB",
    "INR\tB",   "DCR\tB",   "MVI\tB,",  "RLC",
    "???",      "DAD\tB",   "LDAX\tB",  "DCX\tB",
    "INR\tC",   "DCR\tC",   "MVI\tC,",  "RRC",
/*   10          11          12          13                             */
    "???",      "LXI\tD,",  "STAX\tD",  "INX\tD",
    "INR\tD",   "DCR\tD",   "MVI\tD,",  "RAL",
    "???",      "DAD\tD",   "LDAX\tD",  "DCX\tD",
    "INR\tE",   "DCR\tE",   "MVI\tE,",  "RAR",
/*   20          21          22          23                             */
    "RIM",      "LXI\tH,",  "SHLD\t",   "INX\tH",
    "INR\tH",   "DCR\tH",   "MVI\tH,",  "DAA",
    "???",      "DAD\tH",   "LHLD\t",   "DCX\tH",
    "INR\tL",   "DCR\tL",   "MVI\tL,",  "CMA",
/*   30          31          32          33                             */
    "SIM",      "LXI\tSP,", "STA\t",    "INX\tSP",
    "INR\tM",   "DCR\tM",   "MVI\tM,",  "STC",
    "???",      "DAD\tSP",  "LDA\t",    "DCX\tSP",
    "INR\tA",   "DCR\tA",   "MVI\tA,",  "CMC\tA,",
/*   40          41          42          43                             */
    "MOV\tB,B", "MOV\tB,C", "MOV\tB,D", "MOV\tB,E",
    "MOV\tB,H", "MOV\tB,L", "MOV\tB,M", "MOV\tB,A",
    "MOV\tC,B", "MOV\tC,C", "MOV\tC,D", "MOV\tC,E",
    "MOV\tC,H", "MOV\tC,L", "MOV\tC,M", "MOV\tC,A",
/*   50          51          52          53                             */
    "MOV\tD,B", "MOV\tD,C", "MOV\tD,D", "MOV\tD,E",
    "MOV\tD,H", "MOV\tD,L", "MOV\tD,M", "MOV\tD,A",
    "MOV\tE,B", "MOV\tE,C", "MOV\tE,D", "MOV\tE,E",
    "MOV\tE,H", "MOV\tE,L", "MOV\tE,M", "MOV\tE,A",
/*   60          61          62          63                             */
    "MOV\tH,B", "MOV\tH,C", "MOV\tH,D", "MOV\tH,E",
    "MOV\tH,H", "MOV\tH,L", "MOV\tH,M", "MOV\tH,A",
    "MOV\tL,B", "MOV\tL,C", "MOV\tL,D", "MOV\tL,E",
    "MOV\tL,H", "MOV\tL,L", "MOV\tL,M", "MOV\tL,A",
/*   70          71          72          73                             */
    "MOV\tM,B", "MOV\tM,C", "MOV\tM,D", "MOV\tM,E",
    "MOV\tM,H", "MOV\tM,L", "HLT\t",    "MOV\tM,A",
    "MOV\tA,B", "MOV\tA,C", "MOV\tA,D", "MOV\tA,E",
    "MOV\tA,H", "MOV\tA,L", "MOV\tA,M", "MOV\tA,A",
/*   80          81          82          83                             */
    "ADD\tB",   "ADD\tC",   "ADD\tD",   "ADD\tE",
    "ADD\tH",   "ADD\tL",   "ADD\tM",   "ADD\tA",
    "ADC\tB",   "ADC\tC",   "ADC\tD",   "ADC\tE",
    "ADC\tH",   "ADC\tL",   "ADC\tM",   "ADC\tA",
/*   90          91          92          93                             */
    "SUB\tB",   "SUB\tC",   "SUB\tD",   "SUB\tE",
    "SUB\tH",   "SUB\tL",   "SUB\tM",   "SUB\tA",
    "SBB\tB",   "SBB\tC",   "SBB\tD",   "SBB\tE",
    "SBB\tH",   "SBB\tL",   "SBB\tM",   "SBB\tA",
/*   A0          A1          A2          A3                             */
    "ANA\tB",   "ANA\tC",   "ANA\tD",   "ANA\tE",
    "ANA\tH",   "ANA\tL",   "ANA\tM",   "ANA\tA",
    "XRA\tB",   "XRA\tC",   "XRA\tD",   "XRA\tE",
    "XRA\tH",   "XRA\tL",   "XRA\tM",   "XRA\tA",
/*   B0          B1          B2          B3                             */
    "ORA\tB",   "ORA\tC",   "ORA\tD",   "ORA\tE",
    "ORA\tH",   "ORA\tL",   "ORA\tM",   "ORA\tA",
    "CMP\tB",   "CMP\tC",   "CMP\tD",   "CMP\tE",
    "CMP\tH",   "CMP\tL",   "CMP\tM",   "CMP\tA",
/*   C0          C1          C2          C3                             */
    "RNZ",      "POP\tB",   "JNZ\t",    "JMP\t",
    "CNZ\t",    "PUSH\tB",  "ADI\t",    "RST\t0",
    "RZ",       "RET",      "JZ\t",     "???",
    "CZ\t",     "CALL\t",   "ACI\t",    "RST\t1",
/*   D0          D1          D2          D3                             */
    "RNC",      "POP\tD",   "JNC\t",    "OUT\t",
    "CNC\t",    "PUSH\tD",  "SUI\t",    "RST\t2",
    "RC",       "???",      "JC\t",     "IN\t",
    "CC\t",     "???",      "SBI\t",    "RST\t3",
/*   E0          E1          E2          E3                             */
    "RPO",      "POP\tH",   "JPO\t",    "XTHL",
    "CPO\t",    "PUSH\tH",  "ANI\t",    "RST\t4",
    "RPE",      "PCHL",     "JPE\t",    "XCHG",
    "CPE\t",    "???",      "XRI\t",    "RST\t5",
/*   F0          F1          F2          F3                             */
    "RP",       "POP\tPSW", "JP\t",     "DI",
    "CP\t",     "PUSH\tPSW","ORI\t",    "RST\t6",
    "RM",       "SPHL",     "JM\t",     "EI",
    "CM\t",     "???",      "CPI\t",    "RST\t7"
   };

char    Op_Code_Size[]  =
   {
/*  0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F         */
    1,  3,  1,  1,  1,  1,  2,  1,  1,  1,  1,  1,  1,  1,  2,  1,  /* 00 */
    1,  3,  1,  1,  1,  1,  2,  1,  1,  1,  1,  1,  1,  1,  2,  1,  /* 10 */
    1,  3,  3,  1,  1,  1,  2,  1,  1,  1,  3,  1,  1,  1,  2,  1,  /* 20 */
    1,  3,  3,  1,  1,  1,  2,  1,  1,  1,  3,  1,  1,  1,  2,  1,  /* 30 */

    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 40 */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 50 */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 60 */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 70 */

    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 80 */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* 90 */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* A0 */
    1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  /* B0 */

    1,  1,  3,  3,  1,  1,  2,  1,  1,  1,  3,  1,  3,  3,  2,  1,  /* C0 */
    1,  1,  3,  2,  3,  1,  2,  1,  1,  1,  3,  2,  3,  1,  2,  1,  /* D0 */
    1,  1,  3,  1,  3,  1,  2,  1,  1,  1,  3,  1,  3,  1,  2,  1,  /* E0 */
    1,  1,  3,  1,  3,  1,  2,  1,  1,  1,  3,  1,  3,  1,  2,  1   /* F0 */
   };

/*  File extension utilities. */

void Extend( out, in, ext )     /* create out-file name with extension */
char *out, *in, *ext;
{
strcpy( out, in );
if ( strchr( out, '.' )  ==  NULL )
  {
  strcat( out, "." );
  strcat( out, ext );
  }
}

void Strip( out, in )   /* create output file withOUT extension */
char out[], in[];
{
char *p;

strcpy( out, in );
if ( (p = strchr( out, '.' ))  !=  NULL )
  *p  = (char)0;
}

int htoi( char s[] )    /* convert hex string into integer */
{
char HexChar[] = { "0123456789abcdefABCDEF" };
char c;

int i, j, k;
k = 0;
for ( i = 0; i < strlen(s); i++ )
  {
  c = s[i];
  j = strchr( HexChar,c ) - HexChar;    /* convert ascii to number */
  if ( j < 0 ) return (0);              /* invalid hex character */
  if ( j > 15 ) j -= 6;                 /* convert ABCDEF to abcdef */
  k <<= 4;
  k |= j;
  }
return (k);
}

int main( argc, argv )  /* - - - - - - - - mainline - - - - - - - - */
int argc;
char *argv[];
{
FILE *In, *Out;
int c, op, pc,i,j,m,k,ptr;

pc  = 0;        /* program counter starting address default = 0000H */

if ( (argv[1][0]  ==  '?')  ||  ( argc  <  2) )
  {
  printf( "\nUsage:  DB8085  binfile  hexoffset                (v1.0rmb)\n" );
  printf( "Converts binary file to an 8085 Assembler Source Code file.\n" );
  printf( "The binfile extension is optional, [BIN default].\n" );
  printf( "Output is to the same file name, with an SRC extension.\n" );
  printf( "Hexadecimal address offset is optional, [0000 default].\n" );
  return  0;                         
  }

Extend( In_File, argv[1], "BIN" );
if ( (In = fopen( In_File, "rb") )  ==  NULL )
  {
  printf( "Cannot open input file %s!\r\n", In_File );
  return  -1;
  }

Strip( Temp, argv[1] );         /* change filename extension for output */
Extend( Out_File, Temp, "SRC" );
if ( (Out = fopen( Out_File, "wt") )  ==  NULL )
  {
  printf( "Cannot open disassembly list file %s!\r\n", Out_File );
  return -2;
  }
if ( argc == 3 )  pc = htoi( argv[2] );  /* get starting address */

for( i = 0; i<2048; i++ )     /* clear label addresses to FFFF */
  Labels[i] = 0xffff;
    
k = 0;
while ( (op = fgetc( In ))  !=  EOF )  /* PASS-1: scan input for labels */
  {
  if ( Op_Code_Size[ op ]  ==  2 )     /* 2-byte op-code */
    fgetc(In);
  if ( Op_Code_Size[ op ]  ==  3 )     /* 3-byte op-code */
    {
    i = fgetc(In);                     /* get next two bytes for addr */
    j = fgetc(In);
    j <<= 8;                           /* shift left 8 bits for msB */
    i |= j;                            /* form 16-bit address */
    for( m = 0 ; m <= k ; m++)
      {                                /* scan symbol table for entry */
      if( i == Labels[m])
      break;                           /* found it ! */
      if( i < Labels[m])
	{                              /* not in table, add it by */
	for( j = k ; j > m ; j-- )     /* pushing down later ones */
	  Labels[j] = Labels[j-1];     /* to make space for it.   */
	Labels[m] = i;                 /* insert symbol in table. */
	k++;                           /* update table limit size */
	break;
	}
      }
    }
  }

rewind( In );   /* rewind for PASS-2 thru binary file */
ptr = 0;        /* reset label table pointer */
fprintf( Out, "\tORG\t%04XH\n",pc );
while ( (op = fgetc( In ))  !=  EOF )
  {
  while( pc > Labels[ptr] )
    fprintf( Out, "; *** WARNING *** Sync error, L%04X\n", Labels[ptr++] );

  if( pc == Labels[ptr] )   /* output with label */
    {
    fprintf( Out, "L%04X:\t%s", pc, Op_Code[ op ] );
    ptr++;
    }
  else fprintf( Out, "\t%s", Op_Code[ op ] );  /* output without label */

/*  One-byte opcode may have operand */
  if ( Op_Code_Size[ op ] == 1 )
    {
    if ( strlen(Op_Code[op]) < 5 )  fprintf( Out,"\t" );
    fprintf( Out,"\t\t;%04X  %02X",pc,op );
    if (op > 32 && op < 127 )  fprintf( Out,"  %c ",op );
    }

/*  Two-byte opcode are followed by a byte parameter. */
        
  if ( Op_Code_Size[ op ] == 2 )
    {
    i = fgetc(In);
    if( i < 0xA0 )
      fprintf( Out, "%02XH", i );
    else    fprintf( Out, "0%02XH", i );   /* output operand in hex */
    fprintf( Out,"\t\t;%04X  %02X %02X",pc,op,i );
    }

/*  Three-byte opcodes are followed by two bytes in reverse order. */

  if ( Op_Code_Size[ op ]  ==  3 )
    {
    i = fgetc(In);
    j = fgetc(In);
    fprintf( Out, "L%02X%02X", j,i );    /* output operand */
    if ( strlen(Op_Code[op]) < 7 )  fprintf( Out,"\t" );
    fprintf( Out,"\t;%04X  %02X %02X %02X",pc,op,i,j );
    }
  fprintf( Out, "\n" );
  pc  += Op_Code_Size[ op ];
  }
if (ptr < k )         /* output remaining labels as EQUs */
  {
  for ( ; ptr <= k; ptr++)
     fprintf( Out,"L%04X\tEQU\t%04XH\n",Labels[ptr],Labels[ptr] );
  }
fprintf( Out, "\tEND\n" );
fclose( In );      /* end of file processing, close all files */
fclose( Out );
}
